@@ -3,7 +3,7 @@
  * pgpool: a language independent connection pool server for PostgreSQL
  * written by Tatsuo Ishii
  *
- * Copyright (c) 2003-2017	PgPool Global Development Group
+ * Copyright (c) 2003-2013	PgPool Global Development Group
  *
  * Permission to use, copy, modify, and distribute this software and
  * its documentation for any purpose and without fee is hereby
@@ -85,6 +85,7 @@ static POOL_INTERNAL_BUFFER *pool_create_buffer(void);
 static void pool_discard_buffer(POOL_INTERNAL_BUFFER *buffer);
 static void pool_add_buffer(POOL_INTERNAL_BUFFER *buffer, void *data, size_t len);
 static void *pool_get_buffer(POOL_INTERNAL_BUFFER *buffer, size_t *len);
+static char *pool_get_buffer_pointer(POOL_INTERNAL_BUFFER *buffer);
 static char *pool_get_current_cache_buffer(size_t *len);
 static size_t pool_get_buffer_length(POOL_INTERNAL_BUFFER *buffer);
 static void pool_check_and_discard_cache_buffer(int num_oids, int *oids);
@@ -92,6 +93,7 @@ static void pool_check_and_discard_cache_buffer(int num_oids, int *oids);
 static void pool_set_memqcache_blocks(int num_blocks);
 static int pool_get_memqcache_blocks(void);
 static void *pool_memory_cache_address(void);
+static void pool_reset_fsmm(size_t size);
 static void *pool_fsmm_address(void);
 static void pool_update_fsmm(POOL_CACHE_BLOCKID blockid, size_t free_space);
 static POOL_CACHE_BLOCKID pool_get_block(size_t free_space);
@@ -109,11 +111,11 @@ static POOL_CACHE_BLOCKID pool_reuse_block(void);
 static void dump_shmem_cache(POOL_CACHE_BLOCKID blockid);
 #endif
 
+static int pool_hash_reset(int nelements);
 static int pool_hash_insert(POOL_QUERY_HASH *key, POOL_CACHEID *cacheid, bool update);
 static uint32 create_hash_key(POOL_QUERY_HASH *key);
 static volatile POOL_HASH_ELEMENT *get_new_hash_element(void);
 static void put_back_hash_element(volatile POOL_HASH_ELEMENT *element);
-static bool is_free_hash_element(void);
 static char *get_relation_without_alias(RangeVar *relation);
 
 /*
@@ -335,7 +337,7 @@ static int pool_fetch_cache(POOL_CONNECTION_POOL *backend, const char *query, ch
 			{
 				pool_error("pool_fetch_cache: memcached_get failed %s", memcached_strerror(memc, rc));
 				/*
-				 * Turn off memory cache support to prevent future erros.
+				 * Turn off memory cache support to prevent future errors.
 				 */
 				pool_config->memory_cache_enabled = 0;
 				/* Behave as if cache not found */
@@ -579,9 +581,28 @@ POOL_STATUS pool_fetch_from_memory_cache(POOL_CONNECTION *frontend,
 		free(qcache);
 
 		/*
-		 * Send a "READY FOR QUERY" if not in extended query.
+		 * If we are doing extended query, wait and discard Sync
+		 * message from frontend. This is necessary to prevent
+		 * receiving Sync message after Sending Ready for query.
 		 */
-		if (!pool_is_doing_extended_query_message() && MAJOR(backend) == PROTO_MAJOR_V3)
+		if (pool_is_doing_extended_query_message())
+		{
+			char kind;
+			int32 len;
+
+			if (pool_flush(frontend))
+				return POOL_END;
+			if (pool_read(frontend, &kind, 1))
+				return POOL_END;
+			pool_debug("pool_fetch_from_memory_cache: expecting sync: %c", kind);
+			if (pool_read(frontend, &len, sizeof(len)))
+				return POOL_END;
+		}
+
+		/*
+		 * send a "READY FOR QUERY"
+		 */
+		if (MAJOR(backend) == PROTO_MAJOR_V3)
 		{
 			signed char state;
 
@@ -591,7 +612,10 @@ POOL_STATUS pool_fetch_from_memory_cache(POOL_CONNECTION *frontend,
 			state = MASTER(backend)->tstate;
 			send_message(frontend, 'Z', 5, (char *)&state);
 		}
-
+		else
+		{
+			pool_write(frontend, "Z", 1);
+		}
 		if (pool_flush(frontend))
 		{
 			return POOL_END;
@@ -707,7 +731,7 @@ bool pool_is_allow_to_cache(Node *node, char *query)
 	SelectContext ctx;
 
 	/*
-	 * If NO QUERY CACHE coment exists, do not cache.
+	 * If NO QUERY CACHE comment exists, do not cache.
 	 */
 	if (!strncasecmp(query, NO_QUERY_CACHE, NO_QUERY_CACHE_COMMENT_SZ))
 		return false;
@@ -1182,7 +1206,6 @@ int pool_get_database_oid_from_dbname(char *dbname)
 	if (status != POOL_CONTINUE)
 	{    
 		pool_debug("pool_discard_oid_maps_by_db: Failed.");
-		free_select_result(res);
 		return 0;
 	}    
 
@@ -1202,7 +1225,7 @@ int pool_get_database_oid_from_dbname(char *dbname)
 /*
  * Add cache id(shmem case) or hash key(memcached case) to table oid
  * map file.  Caller must hold shmem lock before calling this function
- * to avoid file extention conflict among different pgpool child
+ * to avoid file extension conflict among different pgpool child
  * process.
  * As of pgpool-II 3.2, pool_handle_query_cache is responsible for that.
  * (pool_handle_query_cache -> pool_commit_cache -> pool_add_table_oid_map)
@@ -1295,7 +1318,7 @@ static void pool_add_table_oid_map(POOL_CACHEKEY *cachekey, int num_table_oids,
 		/*
 		 * Below was ifdef-out because of a performance reason.
 		 * Looking for duplicate cache entries in a file needed
-		 * unacceptably high cost. So we gave up this and decieded not
+		 * unacceptably high cost. So we gave up this and decided not
 		 * to care about duplicate entries in the file.
 		 */
 #ifdef NOT_USED
@@ -1357,7 +1380,7 @@ static void pool_add_table_oid_map(POOL_CACHEKEY *cachekey, int num_table_oids,
 
 /*
  * Discard all oid maps at pgpool-II startup.
- * This is neccessary for shmem case.
+ * This is necessary for shmem case.
  */
 void pool_discard_oid_maps(void)
 {
@@ -1384,7 +1407,7 @@ void pool_discard_oid_maps_by_db(int dboid)
 /*
  * Reading cache id(shmem case) or hash key(memcached case) from table
  * oid map file according to table_oids and discard cache entries.  If
- * unlink is true, the file will be unlinked after successfull cache
+ * unlink is true, the file will be unlinked after successful cache
  * removal.
  */
 static void pool_invalidate_query_cache(int num_table_oids, int *table_oid, bool unlinkp, int dboid)
@@ -1544,13 +1567,9 @@ static void pool_reset_memqcache_buffer(void)
 	{
 		POOL_TEMP_QUERY_CACHE *cache;
 
-		pool_debug("pool_reset_memqcache_buffer: discard: %p", session_context->query_cache_array);
-
 		pool_discard_query_cache_array(session_context->query_cache_array);
 		session_context->query_cache_array = pool_create_query_cache_array();
 
-		pool_debug("pool_reset_memqcache_buffer: create: %p", session_context->query_cache_array);
-
 		/*
 		 * if the query context is still under use, we cannot discard
 		 * temporary cache.
@@ -1588,7 +1607,7 @@ bool pool_is_shmem_cache(void)
 }
 
 /*
- * Remeber memory cache number of blocks.
+ * Remember memory cache number of blocks.
  */
 static int memqcache_num_blocks;
 static void pool_set_memqcache_blocks(int num_blocks)
@@ -1613,7 +1632,7 @@ static int pool_get_memqcache_blocks(void)
  */
 size_t pool_shared_memory_cache_size(void)
 {
-	int64 num_blocks;
+	int num_blocks;
 	size_t size;
 
 	if (pool_config->memqcache_maxcache > pool_config->memqcache_cache_block_size)
@@ -1627,21 +1646,21 @@ size_t pool_shared_memory_cache_size(void)
 		pool_config->memqcache_cache_block_size;
 	if (num_blocks == 0)
 	{
-		pool_error("pool_shared_memory_cache_size: memqcache_total_size %ld should be greater or equal to memqcache_cache_block_size %d",
+		pool_error("pool_shared_memory_cache_size: memqcache_total_size %d should be greater or equal to memqcache_cache_block_size %d",
 				   pool_config->memqcache_total_size, pool_config->memqcache_cache_block_size);
 		return 0;
 	}
 
-	pool_log("pool_shared_memory_cache_size: number of blocks: %ld", num_blocks);
+	pool_log("pool_shared_memory_cache_size: number of blocks: %d", num_blocks);
 
-	/* Remenber # of blocks */
+	/* Remember # of blocks */
 	pool_set_memqcache_blocks(num_blocks);
 	size = pool_config->memqcache_cache_block_size * num_blocks;
 	return size;
 }
 
 /*
- * Aquire and initialize shared memory cache. This should be called
+ * Acquire and initialize shared memory cache. This should be called
  * only once from pgpool main process at the process staring up time.
  */
 static void *shmem;
@@ -1657,6 +1676,36 @@ int pool_init_memory_cache(size_t size)
 	return 0;
 }
 
+/*
+ * Clear all the shared memory cache and reset FSMM and hash table.
+ */
+void
+pool_clear_memory_cache(void)
+{
+	size_t size;
+#ifdef HAVE_SIGPROCMASK
+	sigset_t oldmask;
+#else
+	int	oldmask;
+#endif
+
+	POOL_SETMASK2(&BlockSig, &oldmask);
+	pool_shmem_lock();
+
+	size = pool_shared_memory_cache_size();
+	memset(shmem, 0, size);
+
+	size = pool_shared_memory_fsmm_size();
+	pool_reset_fsmm(size);
+
+	pool_discard_oid_maps();
+
+	pool_hash_reset(pool_config->memqcache_max_num_cache);
+
+	pool_shmem_unlock();
+	POOL_SETMASK(&oldmask);
+}
+
 /*
  * Return shared memory cache address
  */
@@ -1674,9 +1723,9 @@ static void *pool_memory_cache_address(void)
  * 
  * Free space management map (FSMM) consists of bytes. Each byte
  * corresponds to block id. For example, if you have 1GB cache and
- * block size is 8Kb, number of blocks = 131,072, thus total sizeo of
+ * block size is 8Kb, number of blocks = 131,072, thus total size of
  * FSMM is 128Kb.  Each FSMM entry has value from 0 to 255. Those
- * valus describes total free space in each block.
+ * values describes total free space in each block.
  * For example, if the value is 2, the free space can be between 64
  * bytes and 95 bytes.
  *
@@ -1702,9 +1751,9 @@ size_t pool_shared_memory_fsmm_size(void)
 }
 
 /*
- * Aquire and initialize shared memory cache for FSMM. This should be
+ * Acquire and initialize shared memory cache for FSMM. This should be
  * called after pool_shared_memory_cache_size only once from pgpool
- * main process at the proces staring up time.
+ * main process at the process staring up time.
  */
 static void *fsmm;
 int pool_init_fsmm(size_t size)
@@ -1746,7 +1795,21 @@ static int *pool_fsmm_clock_hand;
  */
 void pool_allocate_fsmm_clock_hand(void)
 {
-	pool_fsmm_clock_hand = pool_shared_memory_create(sizeof(*pool_fsmm_clock_hand));
+	pool_fsmm_clock_hand = pool_shared_memory_create(sizeof(pool_fsmm_clock_hand));
+	*pool_fsmm_clock_hand = 0;
+}
+
+/*
+ * Reset FSMM.
+ */
+static void
+pool_reset_fsmm(size_t size)
+{
+	int encode_value;
+
+	encode_value = POOL_MAX_FREE_SPACE/POOL_FSMM_RATIO;
+	memset(fsmm, encode_value, size);
+
 	*pool_fsmm_clock_hand = 0;
 }
 
@@ -1808,7 +1871,7 @@ static POOL_CACHE_BLOCKID pool_get_block(size_t free_space)
 		return -1;
 	}
 
-	if (free_space > POOL_MAX_FREE_SPACE)
+	if (free_space < 0 || free_space > POOL_MAX_FREE_SPACE)
 	{
 		pool_error("pool_get_block: invalid free space %zd", free_space);
 		return -1;
@@ -1852,13 +1915,13 @@ static void pool_update_fsmm(POOL_CACHE_BLOCKID blockid, size_t free_space)
 		return;
 	}
 
-	if (blockid >= pool_get_memqcache_blocks())
+	if (blockid < 0 || blockid >= pool_get_memqcache_blocks())
 	{
 		pool_error("pool_set_free_space: invalid block id %d", blockid);
 		return;
 	}
 
-	if (free_space > POOL_MAX_FREE_SPACE)
+	if (free_space < 0 || free_space > POOL_MAX_FREE_SPACE)
 	{
 		pool_error("pool_set_free_space: invalid free space %zd", free_space);
 		return;
@@ -1873,7 +1936,7 @@ static void pool_update_fsmm(POOL_CACHE_BLOCKID blockid, size_t free_space)
 
 /*
  * Add item data to shared memory cache.
- * On successfull registration, returns cache id.
+ * On successful registration, returns cache id.
  * The cache id is overwritten by the subsequent call to this function.
  * On error returns NULL.
  */
@@ -1930,22 +1993,12 @@ static POOL_CACHEID *pool_add_item_shmem_cache(POOL_QUERY_HASH *query_hash, char
 	}
 
 	/*
-	 * Initialize the block if neccessary. If no live items are
+	 * Initialize the block if necessary. If no live items are
 	 * remained, we also initialize the block. If there's contiguous
 	 * deleted items, we turn them into free space as well.
 	 */
 	pool_init_cache_block(blockid);
 
-	/*
-	 * Make sure that we have at least one free hash element.
-	 */
-	while (!is_free_hash_element())
-	{
-		/* If not, reuse next victim block */
-		blockid = pool_reuse_block();
-		pool_init_cache_block(blockid);
-	}
-
 	/* Get block address on shmem */
 	p = block_address(blockid);
 	bh = (POOL_CACHE_BLOCK_HEADER *)p;
@@ -1954,7 +2007,7 @@ static POOL_CACHEID *pool_add_item_shmem_cache(POOL_QUERY_HASH *query_hash, char
 	 * Create contiguous free space. We assume that item bodies are
 	 * ordered from bottom to top of the block, and corresponding item
 	 * pointers are ordered from the youngest to the oldest in the
-	 * beggining of the block.
+	 * beginning of the block.
 	 */
 
 	/*
@@ -2060,7 +2113,7 @@ static POOL_CACHEID *pool_add_item_shmem_cache(POOL_QUERY_HASH *query_hash, char
 	 */
 	if (bh->free_bytes < request_size)
 	{
-		/* This shoud not happen */
+		/* This should not happen */
 		pool_error("pool_add_item_shmem_cache: not enough free space. Free space: %d required: %d block id:%d",
 				   bh->free_bytes, request_size, blockid);
 		return NULL;
@@ -2225,8 +2278,8 @@ static POOL_CACHEID *pool_find_item_on_shmem_cache(POOL_QUERY_HASH *query_hash)
 }
 
 /*
- * Delete item data specifed cache id from shmem.
- * On sccessfull deletion, returns 0.
+ * Delete item data specified cache id from shmem.
+ * On successful deletion, returns 0.
  * Other wise return -1.
  * FSMM is also updated.
  */
@@ -2241,7 +2294,7 @@ static int pool_delete_item_shmem_cache(POOL_CACHEID *cacheid)
 	pool_debug("pool_delete_item_shmem_cache: cacheid:%d itemid:%d",
 			   cacheid->blockid, cacheid->itemid);
 
-	if (cacheid->blockid >= pool_get_memqcache_blocks())
+	if (cacheid->blockid < 0 || cacheid->blockid >= pool_get_memqcache_blocks())
 	{
 		pool_error("pool_delete_item_shmem_cache: invalid block id %d",
 				   cacheid->blockid);
@@ -2256,7 +2309,7 @@ static int pool_delete_item_shmem_cache(POOL_CACHEID *cacheid)
 		return -1;
 	}
 
-	if (cacheid->itemid >= bh->num_items)
+	if (cacheid->itemid < 0 || cacheid->itemid >= bh->num_items)
 	{
 		/*
 		 * This could happen if the block is reused.  Since contents
@@ -2294,7 +2347,7 @@ static int pool_delete_item_shmem_cache(POOL_CACHEID *cacheid)
 	/*
 	 * We do NOT count down bh->num_items here. The deleted space will be recycled
 	 * by pool_add_item_shmem_cache(). However, if this is the last item, we can
-	 * recyle whole block.
+	 * recycle whole block.
 	 *
 	 * 2012/4/1: Now we do not pack data in
 	 * pool_add_item_shmem_cache() for performance reason. Also we
@@ -2334,14 +2387,14 @@ static POOL_CACHE_ITEM_HEADER *pool_cache_item_header(POOL_CACHEID *cacheid)
 {
 	POOL_CACHE_BLOCK_HEADER *bh;
 
-	if (cacheid->blockid >= pool_get_memqcache_blocks())
+	if (cacheid->blockid < 0 || cacheid->blockid >= pool_get_memqcache_blocks())
 	{
 		pool_error("pool_cache_item_header: invalid block id %d", cacheid->blockid);
 		return NULL;
 	}
 
 	bh = (POOL_CACHE_BLOCK_HEADER *)block_address(cacheid->blockid);
-	if (cacheid->itemid >= bh->num_items)
+	if (cacheid->itemid < 0 || cacheid->itemid >= bh->num_items)
 	{
 		pool_error("pool_cache_item_header: invalid item id %d", cacheid->itemid);
 		return NULL;
@@ -2358,7 +2411,7 @@ static int pool_init_cache_block(POOL_CACHE_BLOCKID blockid)
 	char *p;
 	POOL_CACHE_BLOCK_HEADER *bh;
 
-	if (blockid >= pool_get_memqcache_blocks())
+	if (blockid < 0 || blockid >= pool_get_memqcache_blocks())
 	{
 		pool_error("pool_init_cache_block: invalid block id %d", blockid);
 		return -1;
@@ -2413,7 +2466,7 @@ static void pool_wipe_out_cache_block(POOL_CACHE_BLOCKID blockid)
 #endif
 
 /*
- * Aquire lock: XXX giant lock
+ * Acquire lock: XXX giant lock
  */
 void pool_shmem_lock(void)
 {
@@ -2526,11 +2579,8 @@ void pool_discard_query_cache_array(POOL_QUERY_CACHE_ARRAY *cache_array)
 	if (!cache_array)
 		return;
 
-	pool_debug("pool_discard_query_cache_array: num_caches: %d", cache_array->num_caches);
-
 	for (i=0;i<cache_array->num_caches;i++)
 	{
-		pool_debug("pool_discard_query_cache_array: i: %d cache: %p", i, cache_array->caches[i]);
 		pool_discard_temp_query_cache(cache_array->caches[i]);
 	}
 	free(cache_array);
@@ -2547,8 +2597,6 @@ static POOL_QUERY_CACHE_ARRAY * pool_add_query_cache_array(POOL_QUERY_CACHE_ARRA
 	if (!cache_array)
 		return cp;
 
-	pool_debug("pool_add_query_cache_array: num_caches: %d cache: %p", cache_array->num_caches, cache);
-
 	if (cache_array->num_caches >= 	cache_array->array_size)
 	{
 		cache_array->array_size += POOL_QUERY_CACHE_ARRAY_ALLOCATE_NUM;
@@ -2583,7 +2631,6 @@ POOL_TEMP_QUERY_CACHE *pool_create_temp_query_cache(char *query)
 		if (!p->query)
 		{
 			pool_error("pool_create_temp_query_cache: strdup failed");
-			free(p);
 			return NULL;
 		}
 		p->buffer = pool_create_buffer();
@@ -2665,7 +2712,7 @@ static void pool_add_temp_query_cache(POOL_TEMP_QUERY_CACHE *temp_cache, char ki
 
 	if ((buflen+data_len+sizeof(int)+1) > pool_config->memqcache_maxcache)
 	{
-		pool_debug("pool_add_temp_query_cache: data size exceeds memqcache_maxcache. current:%zd requested:%zd memq_maxcache:%d",
+		pool_log("pool_add_temp_query_cache: data size exceeds memqcache_maxcache. current:%zd requested:%zd memq_maxcache:%d",
 				 buflen, data_len+sizeof(int)+1, pool_config->memqcache_maxcache);
 		temp_cache->is_exceeded = true;
 		return;
@@ -2695,7 +2742,15 @@ static void pool_add_oids_temp_query_cache(POOL_TEMP_QUERY_CACHE *temp_cache, in
 }
 
 /*
- * Internal buffer management modules
+ * Internal buffer management modules.
+ * Usage:
+ * 1) Create buffer using pool_create_buffer().
+ * 2) Add data to buffer using pool_add_buffer().
+ * 3) Extract (copied) data from buffer using pool_get_buffer().
+ * 4) Optionally you can:
+ *		Obtain buffer length by using pool_get_buffer_length().
+ *		Obtain buffer pointer by using pool_get_buffer_pointer().
+ * 5) Discard buffer using pool_discard_buffer().
  */
 
 /*
@@ -2800,6 +2855,17 @@ static size_t pool_get_buffer_length(POOL_INTERNAL_BUFFER *buffer)
 	return buffer->buflen;
 }
 
+/*
+ * Get internal buffer pointer.
+ */
+static char *pool_get_buffer_pointer(POOL_INTERNAL_BUFFER *buffer)
+{
+	if (buffer == NULL)
+		return NULL;
+
+	return buffer->buf;
+}
+
 /*
  * Get query cache buffer struct of current query context
  */
@@ -2840,7 +2906,7 @@ static char *pool_get_current_cache_buffer(size_t *len)
 
 /*
  * Mark this temporary query cache buffer discarded if the SELECT
- * uses the table oid specied by oids.
+ * uses the table oid specified by oids.
  */
 static void pool_check_and_discard_cache_buffer(int num_oids, int *oids)
 {
@@ -2865,8 +2931,6 @@ static void pool_check_and_discard_cache_buffer(int num_oids, int *oids)
 			continue;
 
 		soids = (int *)pool_get_buffer(cache->oids, &len);
-		if (!soids || !len)
-			continue;
 
 		for(j=0;j<cache->num_oids;j++)
 		{
@@ -2941,23 +3005,6 @@ void pool_handle_query_cache(POOL_CONNECTION_POOL *backend, char *query, Node *n
 					{
 						pool_error("ReadyForQuery: pool_commit_cache failed");
 					}
-					else
-					{
-						/*
-						 * Reset temporary query cache buffer. This is
-						 * necessary if extended query protocol is used and a
-						 * bind/execute message arrives which uses a statement
-						 * created by prior parse message. In this case since
-						 * the temp_cache is not initialized by a parse
-						 * message, messages are added to pre existing temp
-						 * cache buffer. The problem was found in bug#152.
-						 * http://www.pgpool.net/mantisbt/view.php?id=152
-						 */
-						POOL_TEMP_QUERY_CACHE *cache;
-						cache = pool_get_current_cache();
-						pool_discard_temp_query_cache(cache);
-						session_context->query_context->temp_cache = pool_create_temp_query_cache(query);
-					}
 					free(cache_buffer);
 				}
 				pool_shmem_unlock();
@@ -3001,16 +3048,6 @@ void pool_handle_query_cache(POOL_CONNECTION_POOL *backend, char *query, Node *n
 			{
 				session_context->query_cache_array = 
 					pool_add_query_cache_array(session_context->query_cache_array, cache);
-				/*
-				 * Reset temp_cache pointer in the current query
-				 * context so that we don't add the same temp cache to
-				 * the cache array. This is necessary such that case
-				 * when next query is just a "bind message", without
-				 * "parse message". In the case the query context is
-				 * reused and same cache pointer will be added to the
-				 * query_cache_array which we do not want.
-				 */
-				session_context->query_context->temp_cache = NULL;
 			}
 
 			/* Count up temporary SELECT stats */
@@ -3060,10 +3097,8 @@ void pool_handle_query_cache(POOL_CONNECTION_POOL *backend, char *query, Node *n
 			{
 				pool_error("ReadyForQuery: pool_commit_cache failed");
 			}
-			if (oids)
-				free(oids);
-			if (cache_buffer)
-				free(cache_buffer);
+			free(oids);
+			free(cache_buffer);
 		}
 		pool_shmem_unlock();
 		POOL_SETMASK(&oldmask);
@@ -3076,7 +3111,7 @@ void pool_handle_query_cache(POOL_CONNECTION_POOL *backend, char *query, Node *n
 	else		/* Non cache safe queries */
 	{
 		/* Non cachable SELECT */
-		if (node && IsA(node, SelectStmt))
+		if (IsA(node, SelectStmt))
 		{
 			if (state == 'I')
 			{
@@ -3145,14 +3180,6 @@ void pool_handle_query_cache(POOL_CONNECTION_POOL *backend, char *query, Node *n
 					pool_check_and_discard_cache_buffer(num_oids, oids);
 				} 
 			}
-			else if (num_oids == 0)
-			{
-				/* 
-				 * It is also necessary to clear cache buffers in case of
-				 * no oid queries (like BEGIN, CHECKPOINT, VACUUM, etc) too.
-				 */
-				pool_reset_memqcache_buffer();
-			}
 		}
 	}
 }
@@ -3205,7 +3232,7 @@ POOL_QUERY_CACHE_STATS *pool_get_memqcache_stats(void)
 
 /*
  * Reset query cache stats. Caller must lock QUERY_CACHE_STATS_SEM if
- * neccessary.
+ * necessary.
  */
 void pool_reset_memqcache_stats(void)
 {
@@ -3214,8 +3241,8 @@ void pool_reset_memqcache_stats(void)
 }
 
 /*
- * Count up number of successfull SELECTs and returns the number.
- * QUERY_CACHE_STATS_SEM lock is aquired in this function.
+ * Count up number of successful SELECTs and returns the number.
+ * QUERY_CACHE_STATS_SEM lock is acquired in this function.
  */
 long long int pool_stats_count_up_num_selects(long long int num)
 {
@@ -3234,7 +3261,7 @@ long long int pool_stats_count_up_num_selects(long long int num)
 }
 
 /*
- * Count up number of successfull SELECTs in temporary area and returns
+ * Count up number of successful SELECTs in temporary area and returns
  * the number.
  */
 long long int pool_tmp_stats_count_up_num_selects(void)
@@ -3247,7 +3274,7 @@ long long int pool_tmp_stats_count_up_num_selects(void)
 }
 
 /*
- * Return number of successfull SELECTs in temporary area.
+ * Return number of successful SELECTs in temporary area.
  */
 long long int pool_tmp_stats_get_num_selects(void)
 {
@@ -3258,7 +3285,7 @@ long long int pool_tmp_stats_get_num_selects(void)
 }
 
 /*
- * Reset number of successfull SELECTs in temporary area.
+ * Reset number of successful SELECTs in temporary area.
  */
 void pool_tmp_stats_reset_num_selects(void)
 {
@@ -3270,7 +3297,7 @@ void pool_tmp_stats_reset_num_selects(void)
 
 /*
  * Count up number of SELECTs extracted from cache returns the number.
- * QUERY_CACHE_STATS_SEM lock is aquired in this function.
+ * QUERY_CACHE_STATS_SEM lock is acquired in this function.
  */
 long long int pool_stats_count_up_num_cache_hits(void)
 {
@@ -3370,6 +3397,58 @@ int pool_hash_init(int nelements)
 	return 0;
 }
 
+/*
+ * Reset hash table on shared memory "nelements" is max number of
+ * hash keys. The actual number of hash key is rounded up to power of
+ * 2.
+ */
+static int
+pool_hash_reset(int nelements)
+{
+	size_t size;
+	int nelements2;		/* number of rounded up hash keys */
+	int shift;
+	uint32 mask;
+	POOL_HASH_HEADER hh;
+	int i;
+
+	if (nelements <= 0)
+	{
+		pool_error("pool_hash_clear: invalid nelements:%d", nelements);
+		return -1;
+	}
+
+	/* Round up to power of 2 */
+	shift = 32;
+	nelements2 = 1;
+	do
+	{
+		nelements2 <<= 1;
+		shift--;
+	} while (nelements2 < nelements);
+
+	mask = ~0;
+	mask >>= shift;
+
+	size = (char *)&hh.elements - (char *)&hh + sizeof(POOL_HEADER_ELEMENT)*nelements2;
+	memset((void *)hash_header, 0, size);
+
+	hash_header->nhash = nelements2;
+    	hash_header->mask = mask;
+
+	size = sizeof(POOL_HASH_ELEMENT)*nelements2;
+	memset((void *)hash_elements, 0, size);
+
+	for (i=0;i<nelements2-1;i++)
+	{
+		hash_elements[i].next = (POOL_HASH_ELEMENT *)&hash_elements[i+1];
+	}
+	hash_elements[nelements2-1].next = NULL;
+	hash_free = hash_elements;
+
+	return 0;
+}
+
 /*
  * Search cacheid by MD5 hash key string
  * If found, returns cache id, otherwise NULL.
@@ -3611,20 +3690,12 @@ static void put_back_hash_element(volatile POOL_HASH_ELEMENT *element)
 	element->next = elm;
 }
 
-/*
- * Return true if there's a free hash element.
- */
-static bool is_free_hash_element(void)
-{
-	return hash_free->next != NULL;
-}
-
 /*
  * Returns shared memory cache stats.
  * Subsequent call to this function will break return value
  * because its in static memory.
  * Caller must hold shmem_lock before calling this function.
- * If on mememory query cache is not enabled, all stats are 0.
+ * If on memory query cache is not enabled, all stats are 0.
  */
 POOL_SHMEM_STATS *pool_get_shmem_storage_stats(void)
 {
