@@ -5,7 +5,7 @@
  * pgpool: a language independent connection pool server for PostgreSQL 
  * written by Tatsuo Ishii
  *
- * Copyright (c) 2003-2015	PgPool Global Development Group
+ * Copyright (c) 2003-2012	PgPool Global Development Group
  *
  * Permission to use, copy, modify, and distribute this software and
  * its documentation for any purpose and without fee is hereby
@@ -82,7 +82,7 @@ char query_string_buffer[QUERY_STRING_BUFFER_LEN];
 
 /*
  * query string produced by nodeToString() in simpleQuery().
- * this variable only usefull when enable_query_cache is true.
+ * this variable only useful when enable_query_cache is true.
  */
 char *parsed_query = NULL;
 
@@ -450,10 +450,10 @@ POOL_STATUS SimpleQuery(POOL_CONNECTION *frontend,
 		 * - statement is INSERT
 		 * - either "INSERT LOCK" comment exists or insert_lock directive specified
 		 */
-		if (!RAW_MODE && !STREAM)
+		if (!RAW_MODE)
 		{
 			/*
-			 * If there's only one node to send the commad, there's no
+			 * If there's only one node to send the command, there's no
 			 * point to start a transaction.
 			 */
 			if (pool_multi_node_to_be_sent(query_context))
@@ -701,9 +701,7 @@ POOL_STATUS Execute(POOL_CONNECTION *frontend, POOL_CONNECTION_POOL *backend,
 	/*
 	 * Fetch memory cache if possible
 	 */
-	if (pool_config->memory_cache_enabled && pool_is_likely_select(query) &&
-		!pool_is_writing_transaction() &&
-		(TSTATE(backend, MASTER_SLAVE ? PRIMARY_NODE_ID : REAL_MASTER_NODE_ID) != 'E'))
+	if (pool_config->memory_cache_enabled && pool_is_likely_select(query))
 	{
 		bool foundp;
 		POOL_STATUS status;
@@ -724,7 +722,7 @@ POOL_STATUS Execute(POOL_CONNECTION *frontend, POOL_CONNECTION_POOL *backend,
 		/*
 		 * Add bind message's info to query to search.
 		 */
-		if (query_context->is_cache_safe && bind_msg->param_offset && bind_msg->contents)
+		if (bind_msg->param_offset && bind_msg->contents)
 		{
 			/* Extract binary contents from bind message */
 			char *query_in_bind_msg = bind_msg->contents + bind_msg->param_offset;
@@ -747,7 +745,7 @@ POOL_STATUS Execute(POOL_CONNECTION *frontend, POOL_CONNECTION_POOL *backend,
 				snprintf(hex_str, sizeof(hex_str), (i == 0) ? " %02X" : "%02X", 0xff & query_in_bind_msg[i]);
 				hexlen = strlen(hex_str);
 
-				if ((len+hexlen) >= alloc_len)
+				if ((len+hexlen) > alloc_len)
 				{
 					alloc_len += STR_ALLOC_SIZE;
 					search_query = realloc(search_query, alloc_len);
@@ -764,7 +762,7 @@ POOL_STATUS Execute(POOL_CONNECTION *frontend, POOL_CONNECTION_POOL *backend,
 			query_context->query_w_hex = search_query;
 
 			/*
-			 * When a transaction is comitted, query_context->temp_cache->query is used
+			 * When a transaction is committed, query_context->temp_cache->query is used
 			 * to create md5 hash to search for query cache.
 			 * So overwrite the query text in temp cache to the one with the hex of bind message.
 			 * If not, md5 hash will be created by the query text without bind message, and
@@ -966,11 +964,7 @@ POOL_STATUS Parse(POOL_CONNECTION *frontend, POOL_CONNECTION_POOL *backend,
 
 		node = (Node *) lfirst(list_head(parse_tree_list));
 
-		/* If replication mode, check to see what kind of insert lock is
-		 * neccessary.
-		 */
-		if (REPLICATION)
-			insert_stmt_with_lock = need_insert_lock(backend, stmt, node);
+		insert_stmt_with_lock = need_insert_lock(backend, stmt, node);
 
 		/*
 		 * Start query context
@@ -1189,7 +1183,7 @@ POOL_STATUS Parse(POOL_CONNECTION *frontend, POOL_CONNECTION_POOL *backend,
 		{
 			/*
 			 * Check if other than deadlock error detected.  If so, emit
-			 * log. This is usefull when invalid encoding error occurs. In
+			 * log. This is useful when invalid encoding error occurs. In
 			 * this case, PostgreSQL does not report what statement caused
 			 * that error and make users confused.
 			 */
@@ -1248,12 +1242,11 @@ POOL_STATUS Bind(POOL_CONNECTION *frontend, POOL_CONNECTION_POOL *backend,
 {
 	char *pstmt_name;
 	char *portal_name;
-	char *rewrite_msg = NULL;
+	char *rewrite_msg;
 	POOL_SENT_MESSAGE *parse_msg;
 	POOL_SENT_MESSAGE *bind_msg;
 	POOL_SESSION_CONTEXT *session_context;
 	POOL_QUERY_CONTEXT *query_context;
-	int insert_stmt_with_lock = 0;
 
 	/* Get session context */
 	session_context = pool_get_session_context();
@@ -1295,11 +1288,11 @@ POOL_STATUS Bind(POOL_CONNECTION *frontend, POOL_CONNECTION_POOL *backend,
 	}
 
 	/*
-	 * If the query can ce cached, save its offset of query text in bind message's content.
+	 * If the query can be cached, save its offset of query text in bind message's content.
 	 */
 	if (query_context->is_cache_safe)
 	{
-		bind_msg->param_offset = sizeof(char) * (strlen(portal_name) + strlen(pstmt_name) + 2);
+		bind_msg->param_offset = sizeof(int) + sizeof(char) * (strlen(portal_name) + strlen(pstmt_name));
 	}
 
 	session_context->uncompleted_message = bind_msg;
@@ -1323,47 +1316,25 @@ POOL_STATUS Bind(POOL_CONNECTION *frontend, POOL_CONNECTION_POOL *backend,
 			return POOL_END;
 	}
 
-	/*
-	 * Start a transaction if necessary in replication mode
-	 */
-	if (REPLICATION)
-	{
-		pool_debug("Bind: checking strict query");
-		if (is_strict_query(query_context->parse_tree))
-		{
-			pool_debug("Bind: strict query");
-			start_internal_transaction(frontend, backend, query_context->parse_tree);
-			allow_close_transaction = 1;
-		}
-
-		pool_debug("Bind: checking insert lock");
-		insert_stmt_with_lock = need_insert_lock(backend, query_context->original_query, query_context->parse_tree);
-		if (insert_stmt_with_lock)
-		{
-			pool_debug("Bind: issuing insert lock");
-			/* issue a LOCK command to keep consistency */
-			if (insert_lock(frontend, backend, query_context->original_query, (InsertStmt *)query_context->parse_tree, insert_stmt_with_lock) != POOL_CONTINUE)
-			{
-				pool_query_context_destroy(query_context);
-				return POOL_END;
-			}
-		}
-	}
-
 	pool_debug("Bind: waiting for master completing the query");
 	if (pool_extended_send_and_wait(query_context, "B", len, contents, 1, MASTER_NODE_ID)
 		!= POOL_CONTINUE)
 	{
+		if (rewrite_msg != NULL)
+			/* free(rewrite_msg) */;
 		return POOL_END;
 	}
 
 	if (pool_extended_send_and_wait(query_context, "B", len, contents, -1, MASTER_NODE_ID)
 		!= POOL_CONTINUE)
 	{
+		if (rewrite_msg != NULL)
+			/* free(rewrite_msg) */;
 		return POOL_END;
 	}
-    if(rewrite_msg)
-        free(rewrite_msg);
+
+	if (rewrite_msg != NULL)
+		/* free(rewrite_msg) */;
 
 	return POOL_CONTINUE;
 }
@@ -1456,16 +1427,21 @@ POOL_STATUS Close(POOL_CONNECTION *frontend, POOL_CONNECTION_POOL *backend,
 		msg = pool_get_sent_message('Q', contents+1);
 		if (!msg)
 			msg = pool_get_sent_message('P', contents+1);
-
 		if (!msg)
-			pool_debug("Close: cannot get parse message");
+		{
+			pool_error("Close: cannot get parse message");
+			return POOL_END;
+		}
 	}
 	/* Portal */
 	else if (*contents == 'P')
 	{
 		msg = pool_get_sent_message('B', contents+1);
 		if (!msg)
-			pool_debug("Close: cannot get bind message");
+		{
+			pool_error("Close: cannot get bind message");
+			return POOL_END;
+		}
 	}
 	else
 	{
@@ -1473,23 +1449,6 @@ POOL_STATUS Close(POOL_CONNECTION *frontend, POOL_CONNECTION_POOL *backend,
 		return POOL_END;
 	}
 
-	/*
-	 * As per the postgresql, calling close on non existing portals is not
-	 * an error. So on the same footings we will ignore all such calls and
-	 * return the close complete message to clients with out going to backend
-	 */
-	if (!msg)
-	{
-		int len = htonl(sizeof(len));
-		pool_set_command_success();
-		pool_unset_query_in_progress();
-
-		pool_write(frontend, "3", 1);
-		pool_write_and_flush(frontend, &len, sizeof(len));
-
-		return POOL_CONTINUE;
-	}
-
 	session_context->uncompleted_message = msg;
 	query_context = msg->query_context;
 
@@ -1551,7 +1510,7 @@ POOL_STATUS ReadyForQuery(POOL_CONNECTION *frontend,
 	int i;
 	int len;
 	signed char kind;
-	signed char state = 0;
+	signed char state;
 	POOL_SESSION_CONTEXT *session_context;
 	Node *node = NULL;
 	char *query = NULL;
@@ -1574,6 +1533,7 @@ POOL_STATUS ReadyForQuery(POOL_CONNECTION *frontend,
 	if (session_context->mismatch_ntuples && MAJOR(backend) == PROTO_MAJOR_V3)
 	{
 		int i;
+		signed char state;
 		char kind;
 
 		/*
@@ -1768,7 +1728,7 @@ POOL_STATUS ReadyForQuery(POOL_CONNECTION *frontend,
 				 * If the query was COMMIT/ABORT, clear the history
 				 * that we had a writing command in the transaction
 				 * and forget the transaction isolation level.  This
-				 * is neccessary if succeeding transaction is not an
+				 * is necessary if succeeding transaction is not an
 				 * explicit one.
 				 */
 				else if (is_commit_or_rollback_query(node))
@@ -1809,26 +1769,19 @@ POOL_STATUS ReadyForQuery(POOL_CONNECTION *frontend,
 					 */
 					if (TSTATE(backend, MASTER_SLAVE ? PRIMARY_NODE_ID : REAL_MASTER_NODE_ID) == 'T')
 					{
-						/* However, if the query is "SET TRANSACTION READ ONLY" or its variant,
-						 * don't set it.
-						 */
-						if (!pool_is_transaction_read_only(node))
-						{
-							pool_debug("not SET TRANSACTION READ ONLY");
-							pool_set_writing_transaction();
-						}
+						pool_set_writing_transaction();
 					}
 
 					/*
 					 * If the query was CREATE TEMP TABLE, discard
 					 * temp table relcache because we might have had
 					 * persistent table relation cache which has table
-					 * name as the tem table.
+					 * name as the temp table.
 					 */
 					if (IsA(node, CreateStmt))
 					{
 						CreateStmt *create_table_stmt = (CreateStmt *)node;
-						if (create_table_stmt->relation->istemp)
+						if (create_table_stmt->relation->relpersistence)
 							discard_temp_table_relcache();
 					}
 				}
@@ -1840,14 +1793,13 @@ POOL_STATUS ReadyForQuery(POOL_CONNECTION *frontend,
 				/* If we are doing extended query and the state is after EXECUTE,
 				 * then we can commit cache.
 				 * We check latter condition by looking at query_context->query_w_hex.
-				 * This check is neccessary for certain frame work such as PHP PDO.
+				 * This check is necessary for certain frame work such as PHP PDO.
 				 * It sends Sync message right after PARSE and it produces
 				 * "Ready for query" message from backend.
 				 */
 				if (pool_is_doing_extended_query_message())
 				{
-					if (session_context->query_context &&
-						session_context->query_context->query_state[MASTER_NODE_ID] == POOL_EXECUTE_COMPLETE)
+					if (session_context->query_context->query_state[MASTER_NODE_ID] == POOL_EXECUTE_COMPLETE)
 					{
 						pool_handle_query_cache(backend, session_context->query_context->query_w_hex, node, state);
 						free(session_context->query_context->query_w_hex);
@@ -1856,10 +1808,6 @@ POOL_STATUS ReadyForQuery(POOL_CONNECTION *frontend,
 				}
 				else
 				{
-					if (MAJOR(backend) != PROTO_MAJOR_V3)
-					{
-						state = 'I';	/* XXX I don't think query cache works with PROTO2 protocol */
-					}
 					pool_handle_query_cache(backend, query, node, state);
 				}
 			}
@@ -2133,7 +2081,7 @@ POOL_STATUS CommandComplete(POOL_CONNECTION *frontend, POOL_CONNECTION_POOL *bac
 	rows = extract_ntuples(p);
 
 	/*
-	 * Save number of affcted tuples of master node.
+	 * Save number of affected tuples of master node.
 	 */
 	session_context->ntuples[MASTER_NODE_ID] = rows;
 
@@ -2173,7 +2121,7 @@ POOL_STATUS CommandComplete(POOL_CONNECTION *frontend, POOL_CONNECTION_POOL *bac
 			int n = extract_ntuples(p);
 
 			/*
-			 * Save number of affcted tuples.
+			 * Save number of affected tuples.
 			 */
 			session_context->ntuples[i] = n;
 
@@ -2191,7 +2139,7 @@ POOL_STATUS CommandComplete(POOL_CONNECTION *frontend, POOL_CONNECTION_POOL *bac
 				{
 					/*
 					 * Remember that we have different number of UPDATE/DELETE
-					 * affcted tuples in backends.
+					 * affected tuples in backends.
 					 */
 					session_context->mismatch_ntuples = true;
 				}
@@ -2283,111 +2231,6 @@ POOL_STATUS CommandComplete(POOL_CONNECTION *frontend, POOL_CONNECTION_POOL *bac
 	return POOL_CONTINUE;
 }
 
-POOL_STATUS ParameterDescription(POOL_CONNECTION *frontend,
-								 POOL_CONNECTION_POOL *backend)
-{
-	int len, len1 = 0;
-	char *p = NULL;
-	char *p1 = NULL;
-	int status;
-	int sendlen;
-	int i;
-
-	POOL_SESSION_CONTEXT *session_context;
-	int num_params, send_num_params, num_dmy;
-	char kind = 't';
-
-	session_context = pool_get_session_context();
-	if (!session_context)
-	{
-		pool_error("ParameterDescription: cannot get session context");
-		return POOL_END;
-	}
-
-	/* only in replication mode and rewritten query */
-	if (!REPLICATION || !session_context->query_context->rewritten_query)
-		return SimpleForwardToFrontend('t', frontend, backend);
-
-	/* get number of parameters in original query */
-	num_params = session_context->query_context->num_original_params;
-
-	status = pool_read(MASTER(backend), &len, sizeof(len));
-	if (status < 0)
-	{
-		pool_error("ParameterDescription: error while reading message length");
-		return POOL_END;
-	}
-
-	len = ntohl(len);
-	len -= sizeof(int32);
-	len1 = len;
-
-	/* number of parameters in rewritten query is just discarded */
-	status = pool_read(MASTER(backend), &num_dmy, sizeof(int16));
-	len -= sizeof(int16);
-
-	p = pool_read2(MASTER(backend), len);
-	if (p == NULL)
-		return POOL_END;
-
-	p1 = malloc(len);
-	if (p1 == NULL)
-	{
-		pool_error("ParameterDescription: malloc failed");
-		return POOL_ERROR;
-	}
-	memcpy(p1, p, len);
-
-	for (i=0;i<NUM_BACKENDS;i++)
-	{
-		if (VALID_BACKEND(i) && !IS_MASTER_NODE_ID(i))
-		{
-			status = pool_read(CONNECTION(backend, i), &len, sizeof(len));
-			if (status < 0)
-			{
-				pool_error("ParameterDescription: error while reading message length");
-                free(p1);
-				return POOL_END;
-			}
-
-			len = ntohl(len);
-			len -= sizeof(int32);
-
-			p = pool_read2(CONNECTION(backend, i), len);
-			if (p == NULL)
-            {
-                free(p1);
-				return POOL_END;
-            }
-			if (len != len1)
-			{
-				pool_debug("ParameterDescription: length does not match between backends master(%d) %d th backend(%d) kind:(%c)",
-						   len, i, len1, kind);
-			}
-		}
-	}
-
-	pool_write(frontend, &kind, 1);
-
-	/* send back OIDs of parameters in original query and left are discarded */
-	len = sizeof(int16) + num_params * sizeof(int32);
-	sendlen = htonl(len + sizeof(int32));
-	pool_write(frontend, &sendlen, sizeof(int32));
-
-	send_num_params = htons(num_params);
-	pool_write(frontend, &send_num_params, sizeof(int16));
-
-	if (pool_write_and_flush(frontend, p1, num_params * sizeof(int32)) < 0)
-	{
-		pool_error("ParameterDescription: pool_write_and_flush failed");
-		free(p1);
-		return POOL_END;
-	}
-
-	free(p1);
-	return POOL_CONTINUE;
-}
-
 POOL_STATUS ErrorResponse3(POOL_CONNECTION *frontend,
 						   POOL_CONNECTION_POOL *backend)
 {
@@ -2570,7 +2413,7 @@ POOL_STATUS ProcessFrontendResponse(POOL_CONNECTION *frontend,
 	session_context = pool_get_session_context();
 	if (!session_context)
 	{
-		pool_error("ProcessFrontendResponse: cannot get session context");
+		pool_error("Parse: cannot get session context");
 		return POOL_END;
 	}
 
@@ -2580,7 +2423,7 @@ POOL_STATUS ProcessFrontendResponse(POOL_CONNECTION *frontend,
 	if (pool_read(frontend, &fkind, 1) < 0)
 	{
 		pool_log("ProcessFrontendResponse: failed to read kind from frontend. frontend abnormally exited");
-		return POOL_END_WITH_FRONTEND_ERROR;
+		return POOL_END;
 	}
 
 	pool_debug("ProcessFrontendResponse: kind from frontend %c(%02x)", fkind, fkind);
@@ -2616,7 +2459,7 @@ POOL_STATUS ProcessFrontendResponse(POOL_CONNECTION *frontend,
 	pool_unset_doing_extended_query_message();
 
 	/*
-	 * Alloate buffer and copy the packet contents.  Because inside
+	 * Allocate buffer and copy the packet contents.  Because inside
 	 * these protocol modules, pool_read2 maybe called and modify its
 	 * buffer contents.
 	 */
@@ -2837,10 +2680,6 @@ POOL_STATUS ProcessBackendResponse(POOL_CONNECTION *frontend,
 					pool_unset_query_in_progress();
 				break;
 
-			case 't':	/* ParameterDescription */
-				status = ParameterDescription(frontend, backend);
-				break;
-
 			case 'I':	/* EmptyQueryResponse */
 				status = SimpleForwardToFrontend(kind, frontend, backend);
 				if (pool_is_doing_extended_query_message())
@@ -3077,9 +2916,8 @@ POOL_STATUS CopyDataRows(POOL_CONNECTION *frontend,
 						id = pool_get_id(info, p1);
 						pool_debug("CopyDataRow: copying id: %d", id);
 						free(p1);
-						if (id < 0 || !VALID_BACKEND(id))
+						if (!VALID_BACKEND(id))
 						{
-							pool_error("CopyDataRow: pool_get_id returns invalid id: %d", id);
 							exit(1);
 						}
 						if (pool_write(CONNECTION(backend, id), &kind, 1))
