@@ -6,7 +6,7 @@
  * pgpool: a language independent connection pool server for PostgreSQL 
  * written by Tatsuo Ishii
  *
- * Copyright (c) 2003-2016	PgPool Global Development Group
+ * Copyright (c) 2003-2013	PgPool Global Development Group
  *
  * Permission to use, copy, modify, and distribute this software and
  * its documentation for any purpose and without fee is hereby
@@ -100,7 +100,6 @@ void pool_start_query(POOL_QUERY_CONTEXT *query_context, char *query, int len, N
 		query_context->parse_tree = node;
 		query_context->virtual_master_node_id = my_master_node_id;
 		query_context->is_cache_safe = false;
-		query_context->num_original_params = -1;
 		if (pool_config->memory_cache_enabled)
 			query_context->temp_cache = pool_create_temp_query_cache(query);
 		pool_set_query_in_progress();
@@ -119,7 +118,7 @@ void pool_set_node_to_be_sent(POOL_QUERY_CONTEXT *query_context, int node_id)
 		return;
 	}
 
-	if (node_id < 0 || node_id >= MAX_NUM_BACKENDS)
+	if (node_id < 0 || node_id > MAX_NUM_BACKENDS)
 	{
 		pool_error("pool_set_node_to_be_sent: invalid node id:%d", node_id);
 		return;
@@ -141,7 +140,7 @@ void pool_unset_node_to_be_sent(POOL_QUERY_CONTEXT *query_context, int node_id)
 		return;
 	}
 
-	if (node_id < 0 || node_id >= MAX_NUM_BACKENDS)
+	if (node_id < 0 || node_id > MAX_NUM_BACKENDS)
 	{
 		pool_error("pool_unset_node_to_be_sent: invalid node id:%d", node_id);
 		return;
@@ -172,14 +171,6 @@ void pool_clear_node_to_be_sent(POOL_QUERY_CONTEXT *query_context)
 void pool_setall_node_to_be_sent(POOL_QUERY_CONTEXT *query_context)
 {
 	int i;
-	POOL_SESSION_CONTEXT *sc;
-
-	sc = pool_get_session_context();
-	if (!sc)
-	{
-		pool_error("pool_setall_node_to_be_sent: no session context");
-		return;
-	}
 
 	if (!query_context)
 	{
@@ -191,20 +182,7 @@ void pool_setall_node_to_be_sent(POOL_QUERY_CONTEXT *query_context)
 	{
 		if (private_backend_status[i] == CON_UP ||
 			(private_backend_status[i] == CON_CONNECT_WAIT))
-		{
-			/*
-			 * In streaming replication mode, if the node is not
-			 * primary node nor load balance node, there's no point to
-			 * send query.
-			 */
-			if (pool_config->master_slave_mode &&
-				!strcmp(pool_config->master_slave_sub_mode, MODE_STREAMREP) &&
-				i != PRIMARY_NODE_ID && i != sc->load_balance_node_id)
-			{
-				continue;
-			}
 			query_context->where_to_send[i] = true;
-		}
 	}
 	return;
 }
@@ -250,7 +228,7 @@ bool pool_is_node_to_be_sent(POOL_QUERY_CONTEXT *query_context, int node_id)
 		return false;
 	}
 
-	if (node_id < 0 || node_id >= MAX_NUM_BACKENDS)
+	if (node_id < 0 || node_id > MAX_NUM_BACKENDS)
 	{
 		pool_error("pool_is_node_to_be_sent: invalid node id:%d", node_id);
 		return false;
@@ -296,28 +274,12 @@ int pool_virtual_master_db_node_id(void)
 
 	if (sc->query_context)
 	{
-		int node_id = sc->query_context->virtual_master_node_id;
-
-		if (STREAM)
-		{
-			 /*
-			  * Make sure that virtual_master_node_id is either primary node
-			  * id or load balance node id.  If not, it is likely that
-			  * virtual_master_node_id is not set up yet. Let's use the
-			  * primary node id.
-			  */
-			if (node_id != sc->load_balance_node_id && node_id != PRIMARY_NODE_ID)
-			{
-				node_id = PRIMARY_NODE_ID;
-			}
-		}
-
-		return node_id;
+		return sc->query_context->virtual_master_node_id;
 	}
 
 	/*
 	 * No query context exists.  If in master/slave mode, returns
-	 * primary node if exists.  Oterwise returns my_master_node_id,
+	 * primary node if exists.  Otherwise returns my_master_node_id,
 	 * which represents the last REAL_MASTER_NODE_ID.
 	 */
 	if (MASTER_SLAVE)
@@ -378,14 +340,14 @@ void pool_where_to_send(POOL_QUERY_CONTEXT *query_context, char *query, Node *no
 	else if (MASTER_SLAVE && query_context->is_multi_statement)
 	{
 		/*
-		 * If we are in master/slave mode and we have multi stametemt
+		 * If we are in master/slave mode and we have multi statement
 		 * query, we should send it to primary server only. Otherwise
 		 * it is possible to send a write query to standby servers
 		 * because we only use the first element of the multi
 		 * statement query and don't care about the rest.  Typical
 		 * situation where we are bugged by this is, "BEGIN;DELETE
 		 * FROM table;END". Note that from pgpool-II 3.1.0
-		 * transactional statements such as "BEGIN" is unconditionaly
+		 * transactional statements such as "BEGIN" is unconditionally
 		 * sent to all nodes(see send_to_where() for more details).
 		 * Someday we might be able to understand all part of multi
 		 * statement queries, but until that day we need this band
@@ -519,7 +481,9 @@ void pool_where_to_send(POOL_QUERY_CONTEXT *query_context, char *query, Node *no
 	}
 	else if (REPLICATION || PARALLEL_MODE)
 	{
-		if (is_select_query(node, query))
+		if (pool_config->load_balance_mode &&
+			is_select_query(node, query) &&
+			MAJOR(backend) == PROTO_MAJOR_V3)
 		{
 			/*
 			 * If a writing function call is used or replicate_select is true,
@@ -529,10 +493,18 @@ void pool_where_to_send(POOL_QUERY_CONTEXT *query_context, char *query, Node *no
 			{
 				pool_setall_node_to_be_sent(query_context);
 			}
-			else if (pool_config->load_balance_mode &&
-					 MAJOR(backend) == PROTO_MAJOR_V3 &&
-					 TSTATE(backend, MASTER_NODE_ID) == 'I')
+			/* 
+			 * If (we are outside of an explicit transaction) OR
+			 * (the transaction has not issued a write query yet, AND
+			 *	transaction isolation level is not SERIALIZABLE)
+			 * we might be able to load balance.
+			 */
+			else if (TSTATE(backend, MASTER_NODE_ID) == 'I' ||
+					 (!pool_is_writing_transaction() &&
+					  !pool_is_failed_transaction() &&
+					  pool_get_transaction_isolation() != POOL_SERIALIZABLE))
 			{
+				BackendInfo *bkinfo = pool_get_node_info(session_context->load_balance_node_id);
 				/* load balance */
 				pool_set_node_to_be_sent(query_context,
 										 session_context->load_balance_node_id);
@@ -726,23 +698,20 @@ POOL_STATUS pool_send_and_wait(POOL_QUERY_CONTEXT *query_context,
 
 		if (wait_for_query_response(frontend, CONNECTION(backend, i), MAJOR(backend)) != POOL_CONTINUE)
 		{
-			if (REPLICATION)
-			{
-				/* Cancel current transaction */
-				CancelPacket cancel_packet;
+			/* Cancel current transaction */
+			CancelPacket cancel_packet;
 
-				cancel_packet.protoVersion = htonl(PROTO_CANCEL);
-				cancel_packet.pid = MASTER_CONNECTION(backend)->pid;
-				cancel_packet.key= MASTER_CONNECTION(backend)->key;
-				cancel_request(&cancel_packet);
-			}
+			cancel_packet.protoVersion = htonl(PROTO_CANCEL);
+			cancel_packet.pid = MASTER_CONNECTION(backend)->pid;
+			cancel_packet.key= MASTER_CONNECTION(backend)->key;
+			cancel_request(&cancel_packet);
 
 			return POOL_END;
 		}
 
 		/*
 		 * Check if some error detected.  If so, emit
-		 * log. This is usefull when invalid encoding error
+		 * log. This is useful when invalid encoding error
 		 * occurs. In this case, PostgreSQL does not report
 		 * what statement caused that error and make users
 		 * confused.
@@ -942,7 +911,7 @@ POOL_STATUS pool_extended_send_and_wait(POOL_QUERY_CONTEXT *query_context,
 
 		/*
 		 * Check if some error detected.  If so, emit
-		 * log. This is usefull when invalid encoding error
+		 * log. This is useful when invalid encoding error
 		 * occurs. In this case, PostgreSQL does not report
 		 * what statement caused that error and make users
 		 * confused.
@@ -994,7 +963,7 @@ static POOL_DEST send_to_where(Node *node, char *query)
 		T_ClosePortalStmt,
 		T_ClusterStmt,
 		T_CopyStmt,
-		T_CreateStmt,	/* CREAE TABLE */
+		T_CreateStmt,	/* CREATE TABLE */
 		T_DefineStmt,	/* CREATE AGGREGATE, OPERATOR, TYPE */
 		T_DropStmt,		/* DROP TABLE etc. */
 		T_TruncateStmt,
@@ -1003,7 +972,6 @@ static POOL_DEST send_to_where(Node *node, char *query)
 		T_IndexStmt,	/* CREATE INDEX */
 		T_CreateFunctionStmt,
 		T_AlterFunctionStmt,
-		T_RemoveFuncStmt,
 		/*
 		T_DoStmt,		Our parser does not support yet
 		*/
@@ -1026,9 +994,7 @@ static POOL_DEST send_to_where(Node *node, char *query)
 		T_VariableShowStmt,
 		T_DiscardStmt,
 		T_CreateTrigStmt,
-		T_DropPropertyStmt,
 		T_CreatePLangStmt,
-		T_DropPLangStmt,
 		T_CreateRoleStmt,
 		T_AlterRoleStmt,
 		T_DropRoleStmt,
@@ -1042,12 +1008,9 @@ static POOL_DEST send_to_where(Node *node, char *query)
 		T_AlterRoleSetStmt,
 		T_CreateConversionStmt,
 		T_CreateCastStmt,
-		T_DropCastStmt,
 		T_CreateOpClassStmt,
 		T_CreateOpFamilyStmt,
 		T_AlterOpFamilyStmt,
-		T_RemoveOpClassStmt,
-		T_RemoveOpFamilyStmt,
 		T_PrepareStmt,
 		T_ExecuteStmt,
 		T_DeallocateStmt,		/* DEALLOCATE */
@@ -1064,10 +1027,8 @@ static POOL_DEST send_to_where(Node *node, char *query)
 		T_AlterTSConfigurationStmt,
 		T_CreateFdwStmt,
 		T_AlterFdwStmt,
-		T_DropFdwStmt,
 		T_CreateForeignServerStmt,
 		T_AlterForeignServerStmt,
-		T_DropForeignServerStmt,
 		T_CreateUserMappingStmt,
 		T_AlterUserMappingStmt,
 		T_DropUserMappingStmt,
@@ -1282,10 +1243,11 @@ void where_to_send_deallocate(POOL_QUERY_CONTEXT *query_context, Node *node)
 	DeallocateStmt *d = (DeallocateStmt *)node;
 	POOL_SENT_MESSAGE *msg;
 
-	/* DEALLOCATE ALL? */
+	/* DELLOCATE ALL? */
 	if (d->name == NULL)
 	{
 		pool_setall_node_to_be_sent(query_context);
+		return;
 	}
 	else
 	{
@@ -1297,16 +1259,16 @@ void where_to_send_deallocate(POOL_QUERY_CONTEXT *query_context, Node *node)
 			/* Inherit same map from PREPARE or PARSE */
 			pool_copy_prep_where(msg->query_context->where_to_send,
 								 query_context->where_to_send);
-			return;
 		}
-		/* prepared statement was not found */
-		pool_setall_node_to_be_sent(query_context);
+		return;
 	}
+	/* prepared statement was not found */
+	pool_setall_node_to_be_sent(query_context);
 }
 
 /*
  * Returns parse tree for current query.
- * Preconition: the query is in progress state.
+ * Precondition: the query is in progress state.
  */
 Node *pool_get_parse_tree(void)
 {
@@ -1357,8 +1319,7 @@ bool is_set_transaction_serializable(Node *node)
 		return false;
 
 	if (((VariableSetStmt *)node)->kind == VAR_SET_VALUE &&
-		(!strcmp(((VariableSetStmt *)node)->name, "transaction_isolation") ||
-                 !strcmp(((VariableSetStmt *)node)->name, "default_transaction_isolation")))
+		!strcmp(((VariableSetStmt *)node)->name, "transaction_isolation"))
 	{
 		List *options = ((VariableSetStmt *)node)->args;
 		foreach(list_item, options)
@@ -1534,7 +1495,7 @@ void pool_set_query_state(POOL_QUERY_CONTEXT *query_context, POOL_QUERY_STATE st
 /*
  * Return -1, 0 or 1 according to s1 is "before, equal or after" s2 in terms of state
  * transition order. 
- * The State transiton order is defined as: UNPARSED < PARSE_COMPLETE < BIND_COMPLETE < EXECUTE_COMPLETE
+ * The State transition order is defined as: UNPARSED < PARSE_COMPLETE < BIND_COMPLETE < EXECUTE_COMPLETE
  */
 int statecmp(POOL_QUERY_STATE s1, POOL_QUERY_STATE s2)
 {
@@ -1711,77 +1672,3 @@ void pool_unset_cache_exceeded(void)
 		sc->query_context->temp_cache->is_exceeded = false;
 	}
 }
-
-/*
- * Return true if one of followings is true
- *
- * SET transaction_read_only TO on
- * SET TRANSACTION READ ONLY
- * SET TRANSACTION CHARACTERISTICS AS TRANSACTION READ ONLY
- *
- * Note that if the node is not a variable statement, returns false.
- */
-bool pool_is_transaction_read_only(Node *node)
-{
-	ListCell   *list_item;
-	bool ret = false;
-
-	if (!IsA(node, VariableSetStmt))
-		return ret;
-
-	/*
-	 * SET transaction_read_only TO on
-	 */
-	if (((VariableSetStmt *)node)->kind == VAR_SET_VALUE &&
-		!strcmp(((VariableSetStmt *)node)->name, "transaction_read_only"))
-	{
-		List *options = ((VariableSetStmt *)node)->args;
-		foreach(list_item, options)
-		{
-			A_Const *v = (A_Const *)lfirst(list_item);
-
-			switch (v->val.type)
-			{
-				case T_String:
-					if (!strcasecmp(v->val.val.str, "on") ||
-						!strcasecmp(v->val.val.str, "t") ||
-						!strcasecmp(v->val.val.str, "true"))
-						ret = true;
-					break;
-				case T_Integer:
-					if (v->val.val.ival)
-						ret = true;
-				default:
-					break;
-			}
-		}
-	}
-
-	/*
-	 * SET SESSION CHARACTERISTICS AS TRANSACTION READ ONLY
-	 * SET TRANSACTION READ ONLY
-	 */
-	else if (((VariableSetStmt *)node)->kind == VAR_SET_MULTI &&
-			 (!strcmp(((VariableSetStmt *)node)->name, "TRANSACTION") ||
-			  !strcmp(((VariableSetStmt *)node)->name, "SESSION CHARACTERISTICS")))
-	{
-		List *options = ((VariableSetStmt *)node)->args;
-		foreach(list_item, options)
-		{
-			DefElem *opt = (DefElem *) lfirst(list_item);
-
-			if (!strcmp("transaction_read_only", opt->defname))
-			{
-				bool read_only;
-
-				read_only = ((A_Const *)opt->arg)->val.val.ival;
-				if (read_only)
-				{
-					ret = true;
-					break;
-				}
-			}
-		}
-	}
-	return ret;
-}
