@@ -35,12 +35,12 @@
 #include "pool.h"
 #include "pool_config.h"
 #include "watchdog.h"
+#include "wd_ext.h"
 
 int wd_IP_up(void);
 int wd_IP_down(void);
 int wd_get_cmd(char * buf, char * cmd);
 static int exec_ifconfig(char * path,char * command);
-char *string_replace(const char *string, const char *pattern, const char *replacement);
 
 int
 wd_IP_up(void)
@@ -49,6 +49,9 @@ wd_IP_up(void)
 	char path[WD_MAX_PATH_LEN];
 	char cmd[128];
 
+	if (strlen(pool_config->delegate_IP) == 0)
+		return WD_NG;
+
 	if (WD_List->delegate_ip_flag == 0)
 	{
 		WD_List->delegate_ip_flag = 1;
@@ -72,6 +75,9 @@ wd_IP_down(void)
 	char cmd[128];
 	int i;
 
+	if (strlen(pool_config->delegate_IP) == 0)
+		return WD_NG;
+
 	if (WD_List->delegate_ip_flag == 1)
 	{
 		WD_List->delegate_ip_flag = 0;
@@ -129,7 +135,7 @@ exec_ifconfig(char * path,char * command)
 	int status;
 	char * args[24];
 	int pid, i = 0;
-	char* buf;
+	char buf[256];
 	char *bp, *ep;
 
 	if (pipe(pfd) == -1)
@@ -137,9 +143,8 @@ exec_ifconfig(char * path,char * command)
 		pool_error("exec_ifconfig: pipe open error:%s",strerror(errno));
 		return WD_NG;
 	}
-
-	buf = string_replace(command,"$_IP_$",pool_config->delegate_IP);
-
+	memset(buf,0,sizeof(buf));
+	strlcpy(buf,command,sizeof(buf));
 	bp = buf;
 	while (*bp == ' ')
 	{
@@ -152,7 +157,14 @@ exec_ifconfig(char * path,char * command)
 		{
 			*ep = '\0';
 		}
-		args[i++] = bp;
+		if (!strncmp(bp,"$_IP_$",5))
+		{
+			args[i++] = pool_config->delegate_IP;
+		}
+		else
+		{
+			args[i++] = bp;
+		}
 		if (ep != NULL)
 		{
 			bp = ep +1;
@@ -179,77 +191,26 @@ exec_ifconfig(char * path,char * command)
 	}
 	else
 	{
-		free(buf);
 		close(pfd[1]);
 		for (;;)
 		{
 			int result;
-
-			result = waitpid(pid, &status, 0);
+			result = wait(&status);
 			if (result < 0)
 			{
 				if (errno == EINTR)
 					continue;
-
-				pool_debug("exec_ifconfig: waitpid() failed. reason: %s ", strerror(errno));
-
 				return WD_NG;
 			}
 
 			if (WIFEXITED(status) == 0 || WEXITSTATUS(status) != 0)
-			{
-				pool_debug("exec_ifconfig: '%s' failed. exit status: %d",
-				           command, WEXITSTATUS(status));
 				return WD_NG;
-			}
 			else
 				break;
 		}
 		close(pfd[0]);
 	}
-	pool_debug("exec_ifconfig: '%s' succeeded", command);
-
 	return WD_OK;
 }
 
-/*
- * string_replace()
- * returns the new malloced string after replacing all
- * occurances of pattern in string with replacement string
- */
-char *
-string_replace(const char *string, const char *pattern, const char *replacement)
-{
-	char *tok = NULL;
-	char *newstr = NULL;
-	char *oldstr = NULL;
-	char *head = NULL;
-	size_t pat_len,rep_len;
-
-	newstr = strdup(string);
-	/* bail out if no pattern or replacement is given */
-	if ( pattern == NULL || replacement == NULL )
-		return newstr;
-
-	pat_len = strlen(pattern);
-	rep_len = strlen(replacement);
-
-	head = newstr;
-	while ( (tok = strstr(head,pattern)))
-	{
-		oldstr = newstr;
-		newstr = malloc ( strlen ( oldstr ) - pat_len + rep_len + 1 );
-
-		memcpy(newstr, oldstr, tok - oldstr );
-		memcpy(newstr + (tok - oldstr), replacement, rep_len );
-		memcpy(newstr + (tok - oldstr) + rep_len, tok + pat_len, strlen(oldstr) - pat_len - (tok - oldstr));
-		/* put the string terminator */
-		memset( newstr + strlen (oldstr) - pat_len + rep_len , 0, 1 );
-		/* move back head right after the last replacement */
-		head = newstr + (tok - oldstr) + rep_len;
-		free(oldstr);
-	}
-	return newstr;
-}
-
 
